"""Update app_id column in records table with matching app_id from apps table.

Revision ID: 8
Revises: 7
Create Date: 2024-08-16 12:46:49.510690
"""

from alembic import op
from sqlalchemy.orm.session import Session
from trulens.core.database.orm import make_orm_for_prefix

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

# revision identifiers, used by Alembic.
revision = "8"
down_revision = "7"
branch_labels = None
depends_on = "6"


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    with Session(bind=op.get_bind()) as session:
        orm = make_orm_for_prefix(table_prefix=prefix)

        apps = session.query(orm.AppDefinition).all()
        if tqdm is not None:
            apps = tqdm(apps, desc="Updating app_id in records table")
        for app in apps:
            op.execute(
                f"UPDATE {prefix + 'records'} SET app_id = '{app.app_id}' WHERE app_id = '{app.app_version}'"
            )

        session.commit()
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###

    with Session(bind=op.get_bind()) as session:
        orm = make_orm_for_prefix(table_prefix=prefix)

        apps = session.query(orm.AppDefinition).all()
        for app in apps:
            op.execute(
                f"UPDATE {prefix + 'records'} SET app_id = '{app.app_version}' WHERE app_id = '{app.app_id}'"
            )
        session.commit()
    # ### end Alembic commands ###



"""Python compatibility utilities.

These are mostly type/function aliases to provide consistency across python
versions.

Imports from this module can violate the import modules standard. That is, it is
ok to import the defined names directly as they are replacement to standard
python names.
"""

from __future__ import annotations

from concurrent import futures
import inspect
import queue
import sys
import types
from typing import (
    Any,
    Generic,
    TypeVar,
)
import weakref

import typing_extensions

TypeAliasType = typing_extensions.TypeAliasType
TypeAlias = typing_extensions.TypeAlias
Type = typing_extensions.Type

if sys.version_info >= (3, 11):
    getmembers_static = inspect.getmembers_static
else:

    def getmembers_static(obj, predicate=None):
        """Implementation of inspect.getmembers_static for python < 3.11."""

        if predicate is None:
            predicate = lambda name, value: True

        return [
            (name, value)
            for name in dir(obj)
            if hasattr(obj, name)
            and predicate(name, value := getattr(obj, name))
        ]


if sys.version_info >= (3, 9):
    Future = futures.Future
    """Alias for [concurrent.futures.Future][].

    In python < 3.9, a subclass of [concurrent.futures.Future][] with
    `Generic[A]` is used instead.
    """

    Queue = queue.Queue
    """Alias for [queue.Queue][] .

    In python < 3.9, a subclass of [queue.Queue][] with
    `Generic[A]` is used instead.
    """

    WeakSet = weakref.WeakSet
    """Alias for [weakref.WeakSet][] .

    In python < 3.9, a subclass of [weakref.WeakSet][] with
    `Generic[A]` is used instead.
    """

    ReferenceType = weakref.ReferenceType
    """Alias for [weakref.ReferenceType][] .

    In python < 3.9, a subclass of [weakref.ReferenceType][] with
    `Generic[A]` is used instead.
    """

else:
    # Fake classes which can have type args. In python earlier than 3.9, the
    # classes imported above cannot have type args which is annoying for type
    # annotations. We use these fake ones instead.

    A = TypeVar("A")

    # HACK011
    class Future(Generic[A], futures.Future):
        """Alias for [concurrent.futures.Future][].

        In python < 3.9, a subclass of [concurrent.futures.Future][] with
        `Generic[A]` is used instead.
        """

    # HACK012
    class Queue(Generic[A], queue.Queue):
        """Alias for [queue.Queue][] .

        In python < 3.9, a subclass of [queue.Queue][] with
        `Generic[A]` is used instead.
        """

    class WeakSet(Generic[A], weakref.WeakSet):
        """Alias for [weakref.WeakSet][] .

        In python < 3.9, a subclass of [weakref.WeakSet][] with
        `Generic[A]` is used instead.
        """

    class ReferenceType(Generic[A], weakref.ReferenceType):
        """Alias for [weakref.ReferenceType][] .

        In python < 3.9, a subclass of [weakref.ReferenceType][] with
        `Generic[A]` is used instead.
        """


class EmptyType(type):
    """A type that cannot be instantiated or subclassed."""

    def __new__(mcs, *args, **kwargs):
        raise ValueError("EmptyType cannot be instantiated.")

    def __instancecheck__(cls, __instance: Any) -> bool:
        return False

    def __subclasscheck__(cls, __subclass: Type) -> bool:
        return False


if sys.version_info >= (3, 10):
    NoneType: TypeAlias = types.NoneType
    """Alias for [types.NoneType][] .

    In python < 3.10, it is defined as `type(None)` instead.
    """

else:
    NoneType = type(None)
    """Alias for [types.NoneType][] .

    In python < 3.10, it is defined as `type(None)` instead.
    """



from __future__ import annotations

from abc import abstractmethod
from collections import defaultdict
from enum import Enum
import functools
import importlib
import os
from typing import (
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

import pydantic
from trulens.core.utils import imports as import_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils

T = TypeVar("T")


class Feature(str, Enum):
    """Experimental feature flags.

    Use [TruSession.experimental_enable_feature][trulens.core.session.TruSession.experimental_enable_feature] to enable
    these features:

    Examples:
        ```python
        from trulens.core.session import TruSession
        from trulens.core.experimental import Feature

        session = TruSession()

        session.experimental_enable_feature(Feature.OTEL_TRACING)
        ```
    """

    SIS_COMPATIBILITY = "sis_compatibility"
    OTEL_TRACING = "otel_tracing"
    """OTEL-like tracing.

    !!! Warning
        This changes how wrapped functions are processed. This setting cannot be
        changed after any wrapper is produced.
    """

    @classmethod
    def _repr_all(cls) -> str:
        """Return a string representation of all the feature flags."""

        ret = ""
        for flag in cls:
            ret += f'  {flag.name} = "{flag.value}"\n'

        return ret

    @classmethod
    def _missing_(cls, value: str):
        raise ValueError(
            f"Invalid feature flag `{value}`. Available flags are:\n{cls._repr_all()}"
        )


_FEATURE_SETUPS: Dict[Feature, str] = {
    Feature.OTEL_TRACING: "trulens.experimental.otel_tracing._feature"
}
"""Mapping from experimental flags to their setup class module by name (module
containing _FeatureSetup class).

Using name here as we don't want to import them until they are needed and also
importing them here would result in a circular import.

This is used to check if the optional imports are available before enabling the
feature flags.
"""


class _FeatureSetup(pydantic.BaseModel):
    """Abstract class for utilities that manage experimental features."""

    FEATURE: ClassVar[Feature]
    """The feature flag enabling this feature."""

    REQUIREMENT: ClassVar[import_utils.ImportErrorMessages]
    """The optional imports required to use the feature."""

    @staticmethod
    @abstractmethod
    def assert_optionals_installed() -> None:
        """Assert that the optional requirements for the feature are installed."""

    @staticmethod
    @abstractmethod
    def are_optionals_installed() -> bool:
        """Check if the optional requirements for the feature are installed."""

    @staticmethod
    def assert_can_enable(feature: Feature) -> None:
        """Asserts that the given feature can be enabled.

        This is used to check if the optional imports are available before
        enabling the feature flags.
        """

        if (modname := _FEATURE_SETUPS.get(feature)) is None:
            return

        return _FeatureSetup.load_setup(modname).assert_optionals_installed()

    @staticmethod
    def can_enable(feature: Feature) -> bool:
        """Check if the given feature can be enabled.

        This is used to check if the optional imports are available before
        enabling the feature flags.
        """

        if (modname := _FEATURE_SETUPS.get(feature)) is None:
            return True

        return _FeatureSetup.load_setup(modname).are_optionals_installed()

    @staticmethod
    def load_setup(modname: str) -> Type[_FeatureSetup]:
        """Load the setup class for the given module."""

        mod = importlib.import_module(modname)

        if not hasattr(mod, "_FeatureSetup"):
            raise ImportError(
                f"Module {mod} does not contain a _FeatureSetup class."
            )

        return getattr(mod, "_FeatureSetup")


class _Setting(Generic[T]):
    """A setting that attains some value and can be prevented from further
    changes ("frozen")."""

    def __init__(self, default: T):
        self.value: T = default
        """The stored value."""

        self.frozen_by: Set[str] = set()
        """Set of representations of frames (not in trulens) that have frozen this value.

        If empty, it has not been frozen and can be changed.
        """

    @property
    def is_frozen(self) -> bool:
        """Determine if the setting is frozen."""

        return len(self.frozen_by) > 0

    def set(self, value: Optional[T] = None, freeze: bool = False) -> T:
        """Set/Get the value.

        Set the value first if a value is provided. Make it unchangeable if
        freeze is set.

        Raises:
            ValueError: If the setting has already been frozen and the value is
                different from the current value.
        """

        if value is not None:
            if len(self.frozen_by) > 0 and value != self.value:
                freezing_frames = ""
                for frame_str in self.frozen_by:
                    freezing_frames += f"  {frame_str}\n"

                raise ValueError(
                    f"Feature flag has already been set to {self.value} and cannot be changed. "
                    f"It has been frozen here:\n{freezing_frames}"
                )

            self.value = value

        if freeze:
            frame = python_utils.external_caller_frame(offset=1)
            frame_str = python_utils.code_line(frame, show_source=True)
            # Store representation instead of frame to avoid keeping references
            # that would prevent GC.
            self.frozen_by.add(frame_str)

        return self.value

    def get(self, freeze: bool = False) -> T:
        """Get the value of this setting.

        If freeze is True, freeze the setting so it cannot be changed.
        """

        return self.set(freeze=freeze)

    def freeze(self, value: Optional[T] = None) -> T:
        """Lock the value of this setting.

        If a value is provided, attempt to set the setting first to that value.

        Raises:
            ValueError: If the setting has already been frozen and the value is
                different from the current value.
        """

        return self.set(value, freeze=True)


class _Settings:
    """A collection of settings to enable/disable experimental features.

    A feature can be enabled/disabled and/or frozen so that once it is set, it
    cannot be changed again. Locking is necessary for some features like OTEL
    tracing as once components have been instrumented with old or new tracing,
    the instrumentation cannot be changed.
    """

    def __init__(self):
        self.settings: Dict[Feature, _Setting[bool]] = defaultdict(
            lambda: _Setting(default=False)
        )
        """The settings for the experimental features."""

    def is_frozen(self, flag: Union[Feature, str]) -> bool:
        """Determine if the given setting is frozen."""

        return self.settings[Feature(flag)].is_frozen

    def set(
        self,
        flag: Union[Feature, str],
        *,
        value: Optional[bool] = None,
        freeze: bool = False,
    ) -> bool:
        """Get/Set the given feature flag to the given value.

        Sets the flag to the given value if the value parameter is set. Freezes
        the flag if the freeze parameter is set to True.

        Raises:
            ValueError: If the flag was already frozen to a different value.
        """

        return self.settings[Feature(flag)].set(value=value, freeze=freeze)

    def freeze(
        self, flag: Union[Feature, str], *, value: Optional[bool] = None
    ) -> bool:
        """Lock the given feature flag to the given value.

        If the value is not provided, freeze the flag to its current value.

        Raises:
            ValueError: If the flag has already been frozen to a different value.
        """

        return self.set(flag, value=value, freeze=True)

    def get(self, flag: Union[str, Feature], *, freeze: bool = False) -> bool:
        """Determine the value of the given feature flag."""

        return self.set(flag, freeze=freeze)

    def enable(
        self, flag: Union[Feature, str], *, freeze: bool = False
    ) -> bool:
        """Enable the given feature flag.

        Freeze the flag if the freeze parameter is set to True.

        Raises:
            ValueError: If the flag was already frozen to disabled.
        """

        return self.set(flag, value=True, freeze=freeze)

    def disable(
        self, flag: Union[Feature, str], *, freeze: bool = False
    ) -> bool:
        """Disable the given feature flag.

        Freezes the flag if the freeze parameter is set to True.

        Raises:
            ValueError: If the flag was already frozen to enabled.
        """

        return self.set(flag, value=False, freeze=freeze)

    def set_multiple(
        self,
        flags: Union[
            Iterable[Union[str, Feature]],
            Mapping[Union[str, Feature], bool],
        ],
        freeze: bool = False,
    ):
        """Set multiple feature flags.

        If freeze is set, freeze the flags. If a dictionary is passed, the keys
        are the feature flags and the values are the values to set them to. If a
        list is passed, the flags are set to True.

        Raises:
            ValueError: If any of the flags are already frozen to a different
                value than specified.
        """

        if isinstance(flags, dict):
            for flag, val in flags.items():
                self.set(flag, value=val, freeze=freeze)
        else:
            for flag in flags:
                self.set(flag, value=True, freeze=freeze)


class _WithExperimentalSettings(
    pydantic.BaseModel,
    text_utils.WithIdentString,
):
    """Mixin to add experimental flags and control methods.

    Prints out messages when features are enabled/disabled, frozen, and when a
    setting fails to take up due to earlier freeze.
    """

    _experimental_feature_flags: _Settings = pydantic.PrivateAttr(
        default_factory=_Settings
    )
    """EXPERIMENTAL: Flags to control experimental features."""

    def _experimental_feature(
        self,
        flag: Union[str, Feature],
        *,
        value: Optional[bool] = None,
        freeze: bool = False,
    ) -> bool:
        """Get and/or set the value of the given feature flag.

        Set it first if value is given. Freeze it if `freeze` is set.

        Raises:
            ValueError: If the flag is frozen to a different value.
        """

        # NOTE(piotrm): The printouts are important as we want to make sure the
        # user is aware that they are using a experimental feature.

        flag = Feature(flag)

        was_frozen = self._experimental_feature_flags.is_frozen(flag)

        if value:
            # If the feature has optional requirements, this checks that they
            # are installed and raises an ImportError if not.
            _FeatureSetup.assert_can_enable(flag)

        original_value = self._experimental_feature_flags.get(flag)
        val = self._experimental_feature_flags.set(
            flag, value=value, freeze=freeze
        )
        changed = val != original_value

        if value is not None and changed:
            if val:
                print(
                    f"{text_utils.UNICODE_CHECK} experimental {flag} enabled."
                )
            else:
                print(
                    f"{text_utils.UNICODE_STOP} experimental {flag} disabled."
                )

        if val and freeze and not was_frozen:
            print(
                f"{text_utils.UNICODE_LOCK} experimental {flag} is enabled and cannot be changed."
            )

        return val

    def _experimental_freeze_feature(self, flag: Union[str, Feature]) -> bool:
        """Get and freeze the given feature flag."""

        return self._experimental_feature(flag, freeze=True)

    def experimental_enable_feature(self, flag: Union[str, Feature]) -> bool:
        """Enable the given feature flag.

        Raises:
            ValueError: If the flag is already frozen to disabled.
        """

        return self._experimental_feature(flag, value=True)

    def experimental_disable_feature(self, flag: Union[str, Feature]) -> bool:
        """Disable the given feature flag.

        Raises:
            ValueError: If the flag is already frozen to enabled.
        """

        return self._experimental_feature(flag, value=False)

    def experimental_feature(
        self, flag: Union[str, Feature], *, freeze: bool = False
    ) -> bool:
        """Determine the value of the given feature flag.

        If `freeze` is set, the flag will be frozen to the value returned.
        """

        return self._experimental_feature(flag, freeze=freeze)

    def experimental_set_features(
        self,
        flags: Optional[
            Union[
                Iterable[Union[str, Feature]],
                Mapping[Union[str, Feature], bool],
            ]
        ],
        freeze: bool = False,
    ):
        """Set multiple feature flags.

        If `freeze` is set, the flags will be frozen to the values given.

        Raises:
            ValueError: If any flag is already frozen to a different value than
            provided.
        """
        if os.getenv("TRULENS_OTEL_TRACING", "").lower() in ["1", "true"]:
            self._experimental_feature(
                Feature.OTEL_TRACING, value=True, freeze=True
            )
        if isinstance(flags, dict):
            for flag, val in flags.items():
                self._experimental_feature(flag, value=val, freeze=freeze)
        elif flags is not None:
            for flag in flags:
                self._experimental_feature(flag, value=True, freeze=freeze)

    def _experimental_assert_feature(
        self, flag: Feature, purpose: Optional[str] = None
    ):
        """Raise a ValueError if the given feature flag is not enabled.

        Gives instructions on how to enable the feature flag if error gets
        raised."""

        flag = Feature(flag)

        if purpose is None:
            purpose = "."
        else:
            purpose = f" for {purpose}."

        if not self.experimental_feature(flag):
            raise ValueError(
                f"""Feature flag {flag} is not enabled{purpose} You can enable it in two ways:
    ```python
    from trulens.core.experimental import Feature

    # Enable for this instance when creating it:
    val = {self.__class__.__name__}(experimental_feature_flags=[{flag}]

    # Enable for this instance after it has been created (for features that allows this):
    val.experimental_enable_feature({flag})
    ```
"""
            )

    @staticmethod
    def _experimental_method_override(
        flag: Feature, enabled: T, freeze: bool = False
    ) -> T:
        """Decorator to replace the decorated method with the given one if the
        specified feature is enabled.

        Freezes the flag if the `frozen` parameter is set.

        Example:
            ```python
            class MyClass(_WithExperimentalSettings, ...):

                def my_method_experimental(self, ...): ...

                @MyClass._experimental_method_override(
                    flag=Feature.OTEL_TRACING,
                    enabled=my_method_experimental
                )
                def my_method(self, ...): ...
            ```
        """

        def wrapper(func: T) -> T:
            @functools.wraps(func)
            def wrapped(self: _WithExperimentalSettings, *args, **kwargs):
                if self.experimental_feature(flag, freeze=freeze):
                    return enabled(*args, **kwargs)

                return func(self, *args, **kwargs)

            return wrapped

        return wrapper

    @staticmethod
    def _experimental_method(
        flag: Feature,
        enabled: Callable,
        disabled: Callable,
        freeze: bool = False,
    ) -> Callable:
        """Select between two methods based on the status of a feature flag.

        The selection happens after the method is called. Freezes the flag if
        the `freeze` parameter is set.

        Example:
            ```python
            class MyClass(_WithExperimentalSettings, ...):
                ...
                def my_method_default(self, ...): ...
                def my_method_experimental(self, ...): ...
                ...
                my_method = MyClass._experimental_method(
                    flag=Feature.OTEL_TRACING,
                    enabled=my_method_experimental,
                    disabled=my_method_default
                )
                ```
        """

        @functools.wraps(enabled)  # or disabled
        def wrapper(self: _WithExperimentalSettings, *args, **kwargs):
            if self.experimental_feature(flag, freeze=freeze):
                return enabled(*args, **kwargs)

            return disabled(*args, **kwargs)

        return wrapper



"""
This file contains utility functions specific to certain span types.
"""

from inspect import signature
import logging
from typing import Any, Callable, Dict, List, Optional, Union

from opentelemetry.baggage import get_baggage
from opentelemetry.context import Context
from opentelemetry.trace.span import Span
from opentelemetry.util.types import AttributeValue
from trulens.core.utils import signature as signature_utils
from trulens.otel.semconv.trace import BASE_SCOPE
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)

"""
Type definitions
"""
Attributes = Optional[
    Union[
        Dict[str, Any],
        Callable[
            [
                Optional[Any],
                Optional[Exception],
                List[Any],
                Optional[Dict[str, Any]],
            ],
            Dict[str, Any],
        ],
    ]
]

"""
General utilites for all spans
"""


def validate_selector_name(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """
    Utility function to validate the selector name in the attributes.

    It does the following:
    1. Ensure that the selector name is a string.
    2. Ensure that the selector name is keyed with either the trulens/non-trulens key variations.
    3. Ensure that the selector name is not set in both the trulens and non-trulens key variations.
    """

    result = attributes.copy()

    if (
        SpanAttributes.SELECTOR_NAME_KEY in result
        and SpanAttributes.SELECTOR_NAME in result
    ):
        raise ValueError(
            f"Both {SpanAttributes.SELECTOR_NAME_KEY} and {SpanAttributes.SELECTOR_NAME} cannot be set."
        )

    if SpanAttributes.SELECTOR_NAME in result:
        # Transfer the trulens-namespaced key to the non-trulens-namespaced key.
        result[SpanAttributes.SELECTOR_NAME_KEY] = result[
            SpanAttributes.SELECTOR_NAME
        ]
        del result[SpanAttributes.SELECTOR_NAME]

    if SpanAttributes.SELECTOR_NAME_KEY in result:
        selector_name = result[SpanAttributes.SELECTOR_NAME_KEY]
        if not isinstance(selector_name, str):
            raise ValueError(
                f"Selector name must be a string, not {type(selector_name)}"
            )

    return result


def _convert_to_valid_span_attribute_type(val: Any) -> AttributeValue:
    if isinstance(val, (bool, int, float, str)):
        return val
    if isinstance(val, (list, tuple)):
        for curr_type in [bool, int, float, str]:
            if all([isinstance(curr, curr_type) for curr in val]):
                return val
        return [str(curr) for curr in val]
    return str(val)


def set_span_attribute_safely(
    span: Span,
    key: str,
    value: Any,
) -> None:
    if value is not None:
        span.set_attribute(key, _convert_to_valid_span_attribute_type(value))


def set_string_span_attribute_from_baggage(
    span: Span,
    key: str,
    context: Optional[Context] = None,
) -> None:
    value = get_baggage(key, context)
    if value is not None:
        span.set_attribute(key, str(value))


def validate_attributes(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """
    Utility function to validate span attributes based on the span type.
    """
    if not isinstance(attributes, dict) or any([
        not isinstance(key, str) for key in attributes.keys()
    ]):
        raise ValueError("Attributes must be a dictionary with string keys.")

    if SpanAttributes.SPAN_TYPE in attributes:
        raise ValueError("Span type should not be set in attributes.")

    return validate_selector_name(attributes)
    # TODO: validate Span type attributes.


def set_general_span_attributes(
    span: Span,
    /,
    span_type: SpanAttributes.SpanType,
    context: Optional[Context] = None,
) -> None:
    span.set_attribute(SpanAttributes.SPAN_TYPE, span_type)

    set_string_span_attribute_from_baggage(
        span, SpanAttributes.APP_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.APP_VERSION, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.RECORD_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.TARGET_RECORD_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.EVAL_ROOT_ID, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.EVAL.METRIC_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.RUN_NAME, context
    )
    set_string_span_attribute_from_baggage(
        span, SpanAttributes.INPUT_ID, context
    )


def set_function_call_attributes(
    span: Span,
    ret: Any,
    func_exception: Optional[Exception],
    all_kwargs: Dict[str, Any],
) -> None:
    set_span_attribute_safely(span, SpanAttributes.CALL.RETURN, ret)
    set_span_attribute_safely(span, SpanAttributes.CALL.ERROR, func_exception)
    for k, v in all_kwargs.items():
        set_span_attribute_safely(span, f"{SpanAttributes.CALL.KWARGS}.{k}", v)


def set_user_defined_attributes(
    span: Span,
    *,
    span_type: SpanAttributes.SpanType,
    attributes: Dict[str, Any],
) -> None:
    final_attributes = validate_attributes(attributes)

    for key, value in final_attributes.items():
        span.set_attribute(key, value)
        if (
            key != SpanAttributes.SELECTOR_NAME_KEY
            and SpanAttributes.SELECTOR_NAME_KEY in final_attributes
        ):
            span.set_attribute(
                f"{BASE_SCOPE}.{final_attributes[SpanAttributes.SELECTOR_NAME_KEY]}.{key}",
                value,
            )


"""
RECORD_ROOT SPAN
"""


def get_main_input(func: Callable, args: tuple, kwargs: dict) -> str:
    sig = signature(func)
    bindings = signature(func).bind(*args, **kwargs)
    return signature_utils.main_input(func, sig, bindings)


def set_record_root_span_attributes(
    span: Span,
    /,
    func: Callable,
    args: tuple,
    kwargs: dict,
    ret: Any,
    exception: Optional[Exception],
) -> None:
    set_span_attribute_safely(
        span,
        SpanAttributes.RECORD_ROOT.INPUT,
        get_main_input(func, args, kwargs),
    )
    ground_truth_output = get_baggage(
        SpanAttributes.RECORD_ROOT.GROUND_TRUTH_OUTPUT
    )
    if ground_truth_output:
        set_span_attribute_safely(
            span,
            SpanAttributes.RECORD_ROOT.GROUND_TRUTH_OUTPUT,
            ground_truth_output,
        )
    if exception:
        set_span_attribute_safely(
            span, SpanAttributes.RECORD_ROOT.ERROR, str(exception)
        )
    if ret is not None:
        set_span_attribute_safely(
            span,
            SpanAttributes.RECORD_ROOT.OUTPUT,
            signature_utils.main_output(func, ret),
        )



"""create event table

Revision ID: 10
Revises: 9
Create Date: 2024-12-11 09:32:48.976169
"""

import os

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "10"
down_revision = "9"
branch_labels = None
depends_on = None


def _use_event_table():
    # We only use event table if specifically enabled as it requires the often
    # unsupported JSON type and is for temporary testing purposes anyway.
    return os.getenv("TRULENS_OTEL_TRACING", "").lower() in ["1", "true"]


def upgrade(config) -> None:
    if not _use_event_table():
        return

    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    op.create_table(
        prefix + "events",
        sa.Column("event_id", sa.VARCHAR(length=256), nullable=False),
        sa.Column("record", sa.JSON(), nullable=False),
        sa.Column("record_attributes", sa.JSON(), nullable=False),
        sa.Column("record_type", sa.VARCHAR(length=256), nullable=False),
        sa.Column("resource_attributes", sa.JSON(), nullable=False),
        sa.Column("start_timestamp", sa.TIMESTAMP(), nullable=False),
        sa.Column("timestamp", sa.TIMESTAMP(), nullable=False),
        sa.Column("trace", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("event_id"),
    )


def downgrade(config) -> None:
    if not _use_event_table():
        return

    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    op.drop_table(prefix + "events")



"""Utilities for managing optional requirements of the experimental otel_tracing
feature."""

from trulens.core import experimental
from trulens.core.utils import imports as import_utils

FEATURE = experimental.Feature.OTEL_TRACING
"""Feature controlling the use of this module."""

REQUIREMENT = import_utils.format_import_errors(
    ["opentelemetry-api", "opentelemetry-sdk", "opentelemetry-proto"],
    purpose="otel_tracing experimental feature",
)
"""Optional modules required for the otel_tracing experimental feature."""

with import_utils.OptionalImports(REQUIREMENT) as oi:
    from opentelemetry import sdk
    from opentelemetry import trace


class _FeatureSetup(experimental._FeatureSetup):
    """Utilities for managing the otel_tracing experimental feature."""

    FEATURE = FEATURE
    REQUIREMENT = REQUIREMENT

    @staticmethod
    def assert_optionals_installed():
        """Asserts that the optional requirements for the otel_tracing feature are
        installed."""
        oi.assert_installed([sdk, trace])

    @staticmethod
    def are_optionals_installed():
        """Checks if the optional requirements for the otel_tracing feature are
        installed."""
        return not any(import_utils.is_dummy(m) for m in [sdk, trace])



import uuid

from opentelemetry import trace
from opentelemetry.baggage import get_baggage
from opentelemetry.baggage import remove_baggage
from opentelemetry.baggage import set_baggage
import opentelemetry.context as context_api
from opentelemetry.trace.span import Span
from trulens.experimental.otel_tracing.core.session import TRULENS_SERVICE_NAME
from trulens.otel.semconv.trace import SpanAttributes


class FunctionCallContextManager:
    def __init__(self, span_name: str, is_record_root: bool):
        self.span_name = span_name
        self.is_record_root = is_record_root
        self.span_context_manager = None
        self.token = None

    def __enter__(self) -> Span:
        # Set record_id into context.
        if self.is_record_root:
            record_id = get_baggage(SpanAttributes.RECORD_ID)
            if not record_id:
                record_id = str(uuid.uuid4())
                self.token = context_api.attach(
                    set_baggage(SpanAttributes.RECORD_ID, record_id)
                )
        # Create span.
        self.span_context_manager = (
            trace.get_tracer_provider()
            .get_tracer(TRULENS_SERVICE_NAME)
            .start_as_current_span(name=self.span_name)
        )
        span = self.span_context_manager.__enter__()
        if self.is_record_root:
            span.set_attribute(
                SpanAttributes.SPAN_TYPE, SpanAttributes.SpanType.RECORD_ROOT
            )
        return span

    def __exit__(self, exc_type, exc_val, exc_tb):
        e_ret = None
        # Clean up span.
        try:
            if self.span_context_manager is not None:
                self.span_context_manager.__exit__(exc_type, exc_val, exc_tb)
                self.span_context_manager = None
        except Exception as e_span:
            e_ret = e_span
        # Clean up context.
        try:
            if self.token is not None:
                remove_baggage(SpanAttributes.RECORD_ID)
                context_api.detach(self.token)
                self.token = None
        except Exception as e_context:
            e_ret = e_context if e_ret is None else e_ret
        # Throw any errors we found.
        if e_ret is not None:
            raise e_ret



"""Add groundtruth and dataset tables.

Revision ID: 3
Revises: 2
Create Date: 2024-08-17 15:33:09.416935
"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "3"
down_revision = "2"
branch_labels = None
depends_on = None


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    op.create_table(
        prefix + "dataset",
        sa.Column("dataset_id", sa.VARCHAR(length=256), nullable=False),
        sa.Column("dataset_json", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("dataset_id"),
    )
    op.create_table(
        prefix + "ground_truth",
        sa.Column("ground_truth_id", sa.VARCHAR(length=256), nullable=False),
        sa.Column("dataset_id", sa.Text(), nullable=False),
        sa.Column("ground_truth_json", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("ground_truth_id"),
    )
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    op.drop_table(prefix + "ground_truth")
    op.drop_table(prefix + "dataset")
    # ### end Alembic commands ###



import logging
from typing import Any, Callable, Dict, Optional

from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace import export as otel_export_sdk
from opentelemetry.trace.span import Span
from trulens.core import session as core_session
from trulens.core.database.connector import DBConnector
from trulens.core.utils import python as python_utils
from trulens.core.utils import text as text_utils
from trulens.experimental.otel_tracing.core.exporter.connector import (
    TruLensOtelSpanExporter,
)
from trulens.experimental.otel_tracing.core.span import (
    set_general_span_attributes,
)
from trulens.otel.semconv.trace import SpanAttributes

TRULENS_SERVICE_NAME = "trulens"

logger = logging.getLogger(__name__)


def _set_up_tracer_provider() -> TracerProvider:
    resource = Resource.create({"service.name": TRULENS_SERVICE_NAME})
    provider = TracerProvider(resource=resource)
    trace.set_tracer_provider(provider)

    global_tracer_provider = trace.get_tracer_provider()
    if not isinstance(global_tracer_provider, TracerProvider):
        raise ValueError("Received a TracerProvider of an unexpected type!")
    return global_tracer_provider


def _can_import(to_import: str) -> bool:
    try:
        __import__(to_import)
        return True
    except ImportError:
        return False


class TrulensOtelSpanProcessor(otel_export_sdk.BatchSpanProcessor):
    def on_start(
        self, span: Span, parent_context: Optional[Context] = None
    ) -> None:
        set_general_span_attributes(
            span,
            span_type=SpanAttributes.SpanType.UNKNOWN,
            context=parent_context,
        )


class _TruSession(core_session.TruSession):
    def _validate_otel_exporter(
        self,
        exporter: Optional[otel_export_sdk.SpanExporter],
        connector: DBConnector,
    ) -> otel_export_sdk.SpanExporter:
        if (
            exporter is None
            and _can_import("trulens.connectors.snowflake")
            and _can_import("trulens.connectors.snowflake.otel_exporter")
        ):
            from trulens.connectors.snowflake import SnowflakeConnector
            from trulens.connectors.snowflake.otel_exporter import (
                TruLensSnowflakeSpanExporter,
            )

            if isinstance(connector, SnowflakeConnector):
                exporter = TruLensSnowflakeSpanExporter(connector)
        if not exporter:
            exporter = TruLensOtelSpanExporter(connector)
        if not isinstance(exporter, otel_export_sdk.SpanExporter):
            raise ValueError(
                "Provided exporter must be an OpenTelemetry SpanExporter!"
            )
        self._experimental_otel_exporter = exporter
        return exporter

    def _set_up_otel_exporter(
        self,
        connector: DBConnector,
        exporter: Optional[otel_export_sdk.SpanExporter],
    ):
        logger.info(
            f"{text_utils.UNICODE_CHECK} OpenTelemetry exporter set: {python_utils.class_name(exporter.__class__)}"
        )

        tracer_provider = _set_up_tracer_provider()
        # Setting it here for easy access without having to assert the type every time
        self._experimental_tracer_provider = tracer_provider

        exporter = _TruSession._validate_otel_exporter(
            self, exporter, connector
        )

        self._experimental_otel_span_processor = TrulensOtelSpanProcessor(
            exporter
        )
        tracer_provider.add_span_processor(
            self._experimental_otel_span_processor
        )

    @staticmethod
    def _track_costs_for_module_member(
        module,
        method: str,
        cost_computer: Callable[[Any], Dict[str, Any]],
        span_type: SpanAttributes.SpanType,
    ):
        from trulens.core.otel.instrument import instrument_method

        for cls in dir(module):
            obj = python_utils.safer_getattr(module, cls)
            if (
                obj is not None
                and isinstance(obj, type)
                and hasattr(obj, method)
            ):
                instrument_method(
                    obj,
                    method,
                    span_type=span_type,
                    attributes=lambda ret,
                    exception,
                    *args,
                    **kwargs: cost_computer(ret),
                    must_be_first_wrapper=True,
                )

    @staticmethod
    def _track_costs():
        if _can_import("trulens.providers.cortex.endpoint"):
            from snowflake.cortex._sse_client import SSEClient
            from trulens.core.otel.instrument import instrument_method
            from trulens.providers.cortex.endpoint import CortexCostComputer

            instrument_method(
                SSEClient,
                "events",
                span_type=SpanAttributes.SpanType.UNKNOWN,
                attributes=lambda ret,
                exception,
                *args,
                **kwargs: CortexCostComputer.handle_response(ret),
                must_be_first_wrapper=True,
            )
        if _can_import("trulens.providers.openai.endpoint"):
            import openai
            from openai import resources
            from openai.resources import chat
            from trulens.providers.openai.endpoint import OpenAICostComputer

            for module in [openai, resources, chat]:
                _TruSession._track_costs_for_module_member(
                    module,
                    "create",
                    OpenAICostComputer.handle_response,
                    SpanAttributes.SpanType.UNKNOWN,
                )
        if _can_import("trulens.providers.litellm.endpoint"):
            import litellm
            from trulens.core.otel.instrument import instrument_method
            from trulens.providers.litellm.endpoint import LiteLLMCostComputer

            instrument_method(
                litellm,
                "completion",
                span_type=SpanAttributes.SpanType.GENERATION,
                attributes=lambda ret,
                exception,
                *args,
                **kwargs: LiteLLMCostComputer.handle_response(ret),
                must_be_first_wrapper=True,
            )



"""Threading Utilities."""

from __future__ import annotations

from concurrent import futures
from concurrent.futures import ThreadPoolExecutor as fThreadPoolExecutor
from concurrent.futures import TimeoutError
import contextvars
import inspect
import logging
import threading
from threading import Thread as fThread
from typing import Callable, Optional, TypeVar

from trulens.core._utils.pycompat import Future  # import style exception
from trulens.core.utils import python as python_utils

logger = logging.getLogger(__name__)

DEFAULT_NETWORK_TIMEOUT: float = 10.0  # seconds

A = TypeVar("A")
T = TypeVar("T")


class Thread(fThread):
    """Thread that wraps target with copy of context and stack.

    App components that do not use this thread class might not be properly
    tracked.

    Some libraries are doing something similar so this class may be less and
    less needed over time but is still needed at least for our own uses of
    threads.
    """

    def __init__(
        self,
        name=None,
        group=None,
        target=None,
        args=(),
        kwargs={},
        daemon=None,
    ):
        present_stack = python_utils.WeakWrapper(inspect.stack(0))
        present_context = contextvars.copy_context()

        fThread.__init__(
            self,
            name=name,
            group=group,
            target=present_context.run,
            args=(
                python_utils._future_target_wrapper,
                present_stack,
                target,
                *args,
            ),
            kwargs=kwargs,
            daemon=daemon,
        )


class ThreadPoolExecutor(fThreadPoolExecutor):
    """A ThreadPoolExecutor that keeps track of the stack prior to each thread's
    invocation.

    Apps that do not use this thread pool might not be properly tracked.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def submit(self, fn, /, *args, **kwargs):
        present_stack = python_utils.WeakWrapper(inspect.stack(0))
        present_context = contextvars.copy_context()

        return super().submit(
            present_context.run,
            python_utils._future_target_wrapper,
            present_stack,
            fn,
            *args,
            **kwargs,
        )


# HACK007: Attempt to force other users of Thread to use our version instead.

threading.Thread = Thread

# HACK002: Attempt to make other users of ThreadPoolExecutor use our version
# instead. TODO: this may be redundant with the thread override above.

futures.ThreadPoolExecutor = ThreadPoolExecutor
futures.thread.ThreadPoolExecutor = ThreadPoolExecutor


class TP(metaclass=python_utils.SingletonPerNameMeta):  # "thread processing"
    """Manager of thread pools.

    Singleton.
    """

    MAX_THREADS: int = 128
    """Maximum number of threads to run concurrently."""

    DEBUG_TIMEOUT: Optional[float] = 600.0  # [seconds], None to disable
    """How long to wait (seconds) for any task before restarting it."""

    def __init__(self):
        # Run tasks started with this class using this pool.
        self.thread_pool = ThreadPoolExecutor(
            max_workers=TP.MAX_THREADS, thread_name_prefix="TP.submit"
        )

        # Keep a separate pool for threads whose function is only to wait for
        # the tasks executed in the above pool. Keeping this separate to prevent
        # the deadlock whereas the wait thread waits for a tasks which will
        # never be run because the thread pool is filled with wait threads.
        self.thread_pool_debug_tasks = ThreadPoolExecutor(
            max_workers=TP.MAX_THREADS,
            thread_name_prefix="TP.submit with debug timeout",
        )

        self.completed_tasks = 0
        self.timedout_tasks = 0
        self.failed_tasks = 0

    def _run_with_timeout(
        self,
        func: Callable[[A], T],
        *args,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> T:
        if timeout is None:
            timeout = TP.DEBUG_TIMEOUT

        fut: Future[T] = self.thread_pool.submit(func, *args, **kwargs)

        try:
            res: T = fut.result(timeout=timeout)
            return res

        except TimeoutError as e:
            logger.error(
                "Run of %s in %s timed out after %s second(s).\n%s",
                func.__name__,
                threading.current_thread(),
                TP.DEBUG_TIMEOUT,
                python_utils.code_line(func),
            )

            raise e

        except Exception as e:
            logger.warning(
                "Run of %s in %s failed with: %s",
                {func.__name__},
                threading.current_thread(),
                e,
            )

            raise e

    def submit(
        self,
        func: Callable[[A], T],
        *args,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Future[T]:
        """Submit a task to run.

        Args:
            func: Function to run.

            *args: Positional arguments to pass to the function.

            timeout: How long to wait for the task to complete before killing it.

            **kwargs: Keyword arguments to pass to the function.
        """

        if timeout is None:
            timeout = TP.DEBUG_TIMEOUT

        # TODO(piotrm): need deadlock fixes here. If submit or _submit was called
        # earlier in the stack, do not use a threadpool to evaluate this task
        # and instead create a new thread for it. This prevents tasks in a
        # threadpool adding sub-tasks in the same threadpool which can lead to
        # deadlocks. Alternatively just raise an exception in those cases.

        return self._submit(func, *args, timeout=timeout, **kwargs)

    def _submit(
        self,
        func: Callable[[A], T],
        *args,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Future[T]:
        if timeout is None:
            timeout = TP.DEBUG_TIMEOUT

        # Submit a concurrent tasks to run `func` with the given `args` and
        # `kwargs` but stop with error if it ever takes too long. This is only
        # meant for debugging purposes as we expect all concurrent tasks to have
        # their own retry/timeout capabilities.

        return self.thread_pool_debug_tasks.submit(
            self._run_with_timeout, func, *args, timeout=timeout, **kwargs
        )

    def shutdown(self):
        """Shutdown the pools."""

        self.thread_pool.shutdown()
        self.thread_pool_debug_tasks.shutdown()
